#include <cassert>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <vector>
using namespace std;

const int ANT = 100, DOODLEBUG = 5;
const int X_MAX = 20, Y_MAX = 20;
const int ANT_BREED = 3, DOODLEBUG_BREED = 8;
const int DOODLEBUG_STARVE = 3;
const char ANT_SYMBOL = 'o', DOODLEBUG_SYMBOL = 'X', EMPTY_SYMBOL = '-';
const auto OUT_OF_BOUNDS = nullptr;
const int l1_neighbors[4][2] = {{0,  -1},{0,  1},{-1, 0},{1,  0}}; // Up, down, left and right
typedef int Point2D[2];

class World;

class Organism {
protected:
    int breed_counter;
public:
    Organism() : breed_counter(0) {}
    void surviveBump() { breed_counter++; }
    void breedReset() { breed_counter = 0; }
    virtual char getSymbol() const { return EMPTY_SYMBOL; };
    virtual void move(World *world, int from_x, int from_y) {};
};

class Ant : public Organism {
public:
    char getSymbol() const override { return ANT_SYMBOL; }
    bool shouldBreed() { return breed_counter >= ANT_BREED; }
    void move(World *world, int from_x, int from_y) override;
};

class Doodlebug : public Organism {
    int starve_counter;
public:
    Doodlebug() : Organism(), starve_counter(0) {}
    char getSymbol() const override { return DOODLEBUG_SYMBOL; }
    void starveBump() { starve_counter++; }
    void starveReset() { starve_counter = 0; }
    bool shouldBreed() { return breed_counter >= DOODLEBUG_BREED; }
    bool shouldDie() { return starve_counter >= DOODLEBUG_STARVE; }
    void move(World *world, int from_x, int from_y) override;
};

struct OrganismAndPosition {
    Organism *organism;
    int x;
    int y;
};

class World {
private:
    vector<vector<Organism *>> grid;
public:
    World() : grid(Y_MAX, vector<Organism *>(X_MAX, nullptr)) {}
    Organism *get_grid(int x, int y);
    void set_grid(int x, int y, Organism *creature) { grid[y][x] = creature; }

    void initialize();

    void move_organism(int from_x, int from_y, int to_x, int to_y);
    void sample_air_position(int *row, int *col);
    int *sample_adjacent_air_position(int x, int y);
    int *find_first_adjacent_ant_position(int x, int y);
    void remove_organism(int x, int y);
    vector<OrganismAndPosition> create_organism_process_queue(char type_symbol);
    void move_doodle_bugs();
    void move_ants();
    void cycle_doodlebugs();
    void cycle_ants();
    void step_simulate();
    void print(int round) const;
};

int main() {
    srand(time(nullptr));
    auto world = new World();
    world->initialize();
    world->print(0);
    for (int round = 1;; round++) {
        cout << "Press Enter to Continue";
        cin.ignore(numeric_limits<streamsize>::max(), '\n');
        world->step_simulate();
        world->print(round);
    }
}

void Ant::move(World *world, int from_x, int from_y) {
    int *new_p = world->sample_adjacent_air_position(from_x, from_y);
    if (new_p != nullptr) {
        world->move_organism(from_x, from_y, new_p[0], new_p[1]);
        delete new_p;
    }
    surviveBump();
}

void Doodlebug::move(World *world, int from_x, int from_y) {
    int *new_p = world->find_first_adjacent_ant_position(from_x, from_y);
    if (new_p != nullptr) {
        starveReset();
        world->remove_organism(new_p[0], new_p[1]);
        world->move_organism(from_x, from_y, new_p[0], new_p[1]);
        delete new_p;
    } else {
        // Poor doodlebug is starving
        starveBump();
        int *new_p = world->sample_adjacent_air_position(from_x, from_y);
        if (new_p != nullptr) {
            world->move_organism(from_x, from_y, new_p[0], new_p[1]);
            delete new_p;
        }
    }
    surviveBump();
}

Organism *World::get_grid(int x, int y) {
    assert(x >= 0 && x < X_MAX && y >= 0 && y < Y_MAX);
    return grid[y][x];
}

void World::initialize() {
    for (int ant_size = ANT; ant_size > 0; --ant_size) {
        int row, col;
        sample_air_position(&row, &col);
        set_grid(row, col, new Ant());
    }
    for (int doodlebug_size = DOODLEBUG; doodlebug_size > 0; --doodlebug_size) {
        int row, col;
        sample_air_position(&row, &col);
        set_grid(row, col, new Doodlebug());
    }
}

void World::move_organism(int from_x, int from_y, int to_x, int to_y) {
    auto thing = grid[from_y][from_x];
    grid[from_y][from_x] = grid[to_y][to_x];
    grid[to_y][to_x] = thing;
}

void World::sample_air_position(int *row, int *col) {
    do {
        *row = rand() % X_MAX;
        *col = rand() % Y_MAX;
    } while (get_grid(*row, *col) != nullptr);
}


int *World::sample_adjacent_air_position(int x, int y) {
    auto diff = l1_neighbors[rand() % 4];
    int x_t = x + diff[0];
    int y_t = y + diff[1];
    if (x_t >= 0 && x_t < X_MAX && y_t >= 0 && y_t < Y_MAX) {
        if (get_grid(x_t, y_t) == nullptr) return new int[2]{x_t, y_t};
    }
    return nullptr;
}

int *World::find_first_adjacent_ant_position(int x, int y) {
    for (const auto diff: l1_neighbors) {
        int x_t = x + diff[0];
        int y_t = y + diff[1];
        if (x_t >= 0 && x_t < X_MAX && y_t >= 0 && y_t < Y_MAX) {
            auto thing = get_grid(x_t, y_t);
            if (thing != nullptr && thing->getSymbol() == ANT_SYMBOL) {
                int *result = new int[2]{x_t, y_t};
                return result;
            }
        }
    }
    return nullptr;
}

void World::remove_organism(int x, int y) {
    // The thing at (x, y) is varnished unfortunately
    auto thing = get_grid(x, y);
    assert(thing != nullptr);
    delete thing;
    // Fill in some fresh air here
    set_grid(x, y, nullptr);
}

vector<OrganismAndPosition> World::create_organism_process_queue(char type_symbol) {
    vector<OrganismAndPosition> queue;
    for (int x = 0; x < X_MAX; ++x) {
        for (int y = 0; y < Y_MAX; ++y) {
            auto thing = get_grid(x, y);
            if (thing != nullptr && thing->getSymbol() == type_symbol) {
                queue.push_back(OrganismAndPosition{thing, x, y});
            }
        }
    }
    return queue;
}

void World::move_doodle_bugs() {
    auto queue = create_organism_process_queue(DOODLEBUG_SYMBOL);
    for (const auto item: queue) {
        item.organism->move(this, item.x, item.y);
    }
}

void World::move_ants() {
    auto queue = create_organism_process_queue(ANT_SYMBOL);
    for (const auto item: queue) {
        item.organism->move(this, item.x, item.y);
    }
}

void World::cycle_doodlebugs() {
    for (int x = 0; x < Y_MAX; ++x) {
        for (int y = 0; y < X_MAX; ++y) {
            auto thing = get_grid(x, y);
            if (thing != nullptr && thing->getSymbol() == DOODLEBUG_SYMBOL) {
                // Breed
                if (static_cast<Doodlebug *>(thing)->shouldBreed()) {
                    int *new_p = sample_adjacent_air_position(x, y);
                    if (new_p != nullptr) {
                        // printf("(%d, %d) -> (%d, %d)", x, y, new_p[0], new_p[1]);
                        set_grid(new_p[0], new_p[1], new Doodlebug());
                        delete new_p;
                        static_cast<Doodlebug *>(thing)->breedReset();
                    }
                }
                // Starve cleanups
                if (static_cast<Doodlebug *>(thing)->shouldDie()) {
                    remove_organism(x, y);
                }
            }
        }
    }
}

void World::cycle_ants() {
    auto queue = create_organism_process_queue(ANT_SYMBOL);
    for (const auto item: queue) {
        // Breed
        auto thing = static_cast<Ant *>(item.organism);
        int x = item.x;
        int y = item.y;
        if (thing->shouldBreed()) {
            int *new_p = sample_adjacent_air_position(x, y);
            if (new_p != nullptr) {
                set_grid(new_p[0], new_p[1], new Ant());
                delete new_p;
                thing->breedReset();
            }
        }
    }
}

void World::step_simulate() {
    move_doodle_bugs();
    move_ants();
    cycle_doodlebugs();
    cycle_ants();
}

void World::print(int round) const {
    cout << "World at time " << round << ":" << endl << endl;
    for (const auto &row: grid) {
        for (const auto &cell: row) {
            if (cell == nullptr) {
                cout << EMPTY_SYMBOL << ' ';
            } else if (cell->getSymbol() == ANT_SYMBOL) {
                cout << ANT_SYMBOL << ' ';
            } else {
                cout << DOODLEBUG_SYMBOL << ' ';
            }
        }
        cout << endl;
    }
}